@@ -7,9 +7,12 @@ import os
 import sys
 
 import argparse
+import twisted
 from twisted.python import util
 from twisted.scripts import twistd
 
+import piped
+
 
 def run_trial():
     from twisted.python import modules
@@ -38,7 +41,19 @@ def run_trial():
 def run_piped():
     """ Wrapper around twistd to run piped. """
     parser = _make_parser()
-    args = parser.parse_args()
+
+    # we parse only the known arguments in case we're running a script
+    args, remaining = parser.parse_known_args()
+
+    if args.script:
+        # if we are running a script, the configuration file is the script:
+        args.conf = args.script
+        # replace sys.argv with the arguments that the script might be interested in
+        sys.argv = [sys.argv[0]] + remaining
+    else:
+        # otherwise, we have to re-parse the arguments to disallow unknown
+        # arguments.
+        args = parser.parse_args()
 
     # pass the configuration file as an environment variable that is read in the service.tac
     if args.conf:
@@ -55,10 +70,42 @@ def run_piped():
     _run_twistd_with_config(twistd_config)
 
 
+class VersionAction(argparse.Action):
+    """ We provide our own argparse.Action in order to provide our own formatting. """
+
+    def _find_contrib_versions(self):
+        from twisted.python import modules
+
+        versions = dict()
+
+        contrib_module = modules.getModule('piped.contrib')
+        contrib_module.load()
+
+        for module in contrib_module.iterModules():
+            versions[module.name] = getattr(module.load(), 'version', 'unknown version')
+
+        return versions
+
+    def __call__(self, parser, namespace, values, option_string=None):
+        message = list()
+        message.append('piped (the Piped process daemon) %s, %s' % (piped.version, twisted.version))
+        message.append('')
+        message.append('Contrib packages:')
+        for name, version in sorted(self._find_contrib_versions().items()):
+            message.append('\t%s: %s' % (name, version))
+        message.append('')
+        message.append('Python version: %s' % sys.version.replace('\n', ''))
+        message.append('')
+
+        parser.exit(message=os.linesep.join(message))
+
+
 def _make_parser():
     """ Creates an ArgumentParser to parse the command line options. """
     parser = argparse.ArgumentParser(description='Piped process daemon.')
 
+    parser.add_argument('-v', '--version', action=VersionAction, help='Print version information and exit.', nargs=0)
+
     parser.add_argument('-n', '--nodaemon', action='store_true', help='don\'t daemonize, don\'t use default umask of 0077')
     parser.add_argument('-l', '--logfile',
                         help='log to a specified file, - for stdout. %%d in the filename will be replaced with "[config_basename].log" (default: "-")',
@@ -71,7 +118,12 @@ def _make_parser():
                         help=('Configuration overrides that will be set after the configuration has been loaded. '
                               'This option may be used multiple times in order to set multiple overrides. '
                               'Example: --override=web.admin.port:8080'))
-    parser.add_argument('-c', '--conf', help='Configuration file to use. [PIPED_CONFIGURATION_FILE]', metavar='config_file.yaml', required=True)
+
+    group = parser.add_mutually_exclusive_group(required=True)
+    group.add_argument('-c', '--conf', help='Configuration file to use. [PIPED_CONFIGURATION_FILE]', metavar='config_file.yaml')
+    group.add_argument('-s', '--script',
+                       help=('Script file to use. If this argument is used, any unknown arguments passed to piped '
+                             'will remain on sys.argv and can be consumed by the script.'), metavar='config_file.yaml')
 
     return parser
 
