@@ -9,6 +9,7 @@ import os
 import sys
 import xmlrpclib
 import copy
+import json
 
 from twisted.internet import defer, reactor
 from twisted.application import service
@@ -16,11 +17,6 @@ from twisted.python import failure, filepath, reflect
 
 from piped import log
 
-try:
-    import simplejson as json
-except ImportError:
-    import json
-
 try:
     from collections import OrderedDict
 except ImportError:
@@ -503,6 +499,16 @@ class PipedJSONEncoder(json.JSONEncoder):
         return super(PipedJSONEncoder, self).default(obj)
 
 
+class BatonJSONEncoder(PipedJSONEncoder):
+    """ A custom JSON encoder that falls back to repr() if an exception is raised
+    during JSON encoding. """
+    def default(self, obj):
+        try:
+            return super(BatonJSONEncoder, self).default(obj)
+        except TypeError as te:
+            return repr(obj)
+
+
 class PullFromQueueAndProcessInPipeline(service.Service):
     _waiting_on_queue = None
     _waiting_on_pipeline = None
